from genologics.lims import Lims
from pydantic import BaseModel
from cg_lims.status_db_api import StatusDBAPI


class EPPContextObject(BaseModel):
    lims: Lims
    status_db: StatusDBAPI

    class Config:
        arbitrary_types_allowed = True


class APIContextObject(EPPContextObject):
    host: str
    port: int



from genologics.lims import Lims
from genologics.entities import Artifact, Sample
from pathlib import Path

import pytest
from click.testing import CliRunner

import threading
import time

from limsmock.server import run_server

PORT = 8000
HOST = "127.0.0.1"


def server(file_name: str):
    """Starting up a server based on file_path"""

    file_path = f"tests/fixtures/{file_name}"
    thread = threading.Thread(
        target=run_server,
        args=(
            file_path,
            HOST,
            PORT,
        ),
    )
    thread.daemon = True
    thread.start()
    time.sleep(0.1)


@pytest.fixture
def lims() -> Lims:
    """Get genologics lims instance"""

    return Lims(f"http://{HOST}:{PORT}", "dummy", "dummy")


@pytest.fixture
def artifact_1(lims) -> Artifact:
    """Basic artifact with id 1. Containing no udfs.
    Related to sample_1."""

    artifact = Artifact(lims, id="1")

    return artifact


@pytest.fixture
def sample_1(lims) -> Sample:
    """Basic sample with id S1. Containing no udfs.
    Related to artifact_1."""

    sample = Sample(lims, id="S1")

    return sample


@pytest.fixture
def artifact_2(lims) -> Artifact:
    """Basic artifact with id 2. Containing no udfs.
    Related to sample_2."""

    artifact = Artifact(lims, id="2")

    return artifact


@pytest.fixture
def sample_2(lims) -> Sample:
    """Basic sample with id S2. Containing no udfs.
    Related to artifact_2."""

    sample = Sample(lims, id="S2")

    return sample


@pytest.fixture
def config() -> Path:
    """Get file path to config"""

    return Path("tests/fixtures/config.yaml")


@pytest.fixture
def enzymatic_fragmentation_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/Enzymatic_fragmentation"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def kapa_library_preparation_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/KAPA_Library_Preparation"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def kapa_library_preparation_file_missing_udf() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/KAPA_Library_Preparation_missing_udf"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def hamilton_normalization_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/hamilton_normalization.txt"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def hamilton_sars_cov2_pooling_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/sars-cov2-hamilton-prep-pooling.txt"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def hamilton_sars_cov2_indexing_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/sars-cov2-hamilton-prep-indexing.txt"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture(name="cli_runner")
def fixture_cli_runner() -> CliRunner:
    """Create a CliRunner"""
    runner = CliRunner()
    return runner



#!/usr/bin/env python

import logging
import sys
from typing import List

import click
from genologics.entities import Artifact, Process, Stage
from genologics.lims import Lims

from cg_lims import options
from cg_lims.exceptions import DuplicateSampleError, LimsError, MissingArtifactError
from cg_lims.get.artifacts import filter_artifacts, get_artifacts, get_latest_artifact
from cg_lims.put.queue import queue_artifacts

LOG = logging.getLogger(__name__)


def get_artifacts_to_requeue(
    lims: Lims, rerun_arts: List[Artifact], process_type: List[str]
) -> List[Artifact]:
    """Get input artifacts to define step (output artifacts of sort step)
    Args:
        lims: Lims
        rerun_arts: List # Artifacts with Rerun flag
        process_type: List[str] # Name of step(s) before the requeue step
    """

    artifacts_to_requeue = []
    break_rerun = False
    for art in rerun_arts:
        representative_sample_id = art.samples[0].id  ## hantera med if samples..
        try:
            requeue_art = get_latest_artifact(lims, representative_sample_id, process_type)
        except MissingArtifactError as e:
            LOG.warning(e.message)
            break_rerun = True
            continue
        artifacts_to_requeue.append(requeue_art)
    if break_rerun:
        raise MissingArtifactError("Issues finding artifacts to requeue. See log")
    return list(set(artifacts_to_requeue))


def check_same_sample_in_many_rerun_pools(rerun_arts: List[Artifact]) -> None:
    """Check that the same sample does not occure in more than one of the pools to rerun."""

    all_samples = []

    for art in rerun_arts:
        all_samples += art.samples
    for s in set(all_samples):
        all_samples.remove(s)
    duplicate_samples = list(set(all_samples))
    if duplicate_samples:
        raise DuplicateSampleError(
            f"Waring same sample in many pools: {' ,'.join([sample.id for sample in duplicate_samples])}"
        )


@click.command()
@options.workflow_id(help="Destination workflow id.")
@options.stage_id(help="Destination stage id.")
@options.process_type(
    help="The name(s) of the process type(s) before the requeue step. Fetching artifact to requeue from here."
)
@options.udf(help="UDF that will tell wich artifacts to move.")
@click.pass_context
def rerun_samples(ctx, workflow_id, stage_id, udf, process_type):
    """Script to requeue samples for sequencing."""
    process = ctx.obj["process"]
    lims = ctx.obj["lims"]
    artifacts = get_artifacts(process, False)
    rerun_arts = filter_artifacts(artifacts, udf, True)
    if rerun_arts:
        try:
            artifacts_to_requeue = get_artifacts_to_requeue(lims, rerun_arts, process_type)
            check_same_sample_in_many_rerun_pools(artifacts_to_requeue)
            queue_artifacts(lims, artifacts_to_requeue, workflow_id, stage_id)
            click.echo("Artifacts have been queued.")
        except LimsError as e:
            sys.exit(e.message)



from typing import List


def unique_list_of_ids(entity_list: list) -> List[str]:
    """Arg: entity_list: list of any type of genologics entity.
    Retruning unique list of entity ids."""

    return list(set([e.id for e in entity_list]))



import logging
import sys
from typing import List

import click
from genologics.entities import Artifact

from cg_lims.exceptions import LimsError, LowAmountError
from cg_lims.get.artifacts import get_artifacts
from cg_lims.objects import Pool

LOG = logging.getLogger(__name__)


# The avalible volume of a sample is allways 15.
AVALIBLE_SAMPLE_VOLUME = 15

# A sample should not have les than 187.5 ng in the pool
MINIMUM_SAMPLE_AMOUNT = 187.5


class TwistPool(Pool):
    def __init__(self, artifact: Artifact):
        super().__init__(artifact)
        self.qc_flag = "PASSED"
        self.total_volume = 0
        self.amount_fail = False

    def calculate_amount_and_volume(self):
        """Perform calcualtions for the input artifacts to the pool.
        Based on the udf 'Reads missing (M)', the sample is given an
        equivalent proportion in the pool.
        """

        for art in self.artifact.input_artifact_list():
            reads = art.samples[0].udf.get("Reads missing (M)")
            concentration = art.udf.get("Concentration")
            total_amount = self.artifact.udf.get("Total Amount (ng)")
            if None in [reads, concentration, total_amount]:
                self.qc_flag = "FAILED"
                self.amount_fail = True
                continue
            fract_of_pool = reads / float(self.total_reads_missing)
            amount_to_pool = total_amount * fract_of_pool
            vol = amount_to_pool / concentration
            if vol > AVALIBLE_SAMPLE_VOLUME:
                vol = AVALIBLE_SAMPLE_VOLUME
            self.total_volume += vol
            art.udf["Amount taken (ng)"] = amount_to_pool
            art.udf["Volume of sample (ul)"] = vol
            art.put()
            if (
                amount_to_pool > art.udf.get("Amount (ng)")
                or amount_to_pool < MINIMUM_SAMPLE_AMOUNT
            ):
                self.qc_flag = "FAILED"
                self.amount_fail = True


def calculate_volumes_for_pooling(pools: List[Artifact]):
    """Perform calculations for each pool in the step.
    All pools in the plate must have the same volume to not overdry.
    Wahter is added to each pool for them to all get the same volume."""

    amount_failed_for_some_pool = False
    all_volumes = []

    for pool_art in pools:
        twist_pool = TwistPool(pool_art)
        twist_pool.get_total_reads_missing()
        twist_pool.calculate_amount_and_volume()
        pool_art.qc_flag = twist_pool.qc_flag
        if twist_pool.amount_fail:
            amount_failed_for_some_pool = True
        if twist_pool.total_volume:
            pool_art.udf["Total Volume (ul)"] = twist_pool.total_volume
        all_volumes.append(twist_pool.total_volume)

    for pool_art in pools:
        if pool_art.udf.get("Total Volume (ul)"):
            pool_art.udf["Volume H2O (ul)"] = max(all_volumes) - pool_art.udf.get(
                "Total Volume (ul)"
            )
        pool_art.put()

    if amount_failed_for_some_pool:
        raise LowAmountError(
            "Input amount low for samples in some pool or udfs missing. Generate placement map for more info."
        )


@click.command()
@click.pass_context
def twist_pool(ctx):
    """Calculates volumes for pools in a plate before hybridization."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    try:
        pools = get_artifacts(process=process, input=False)
        calculate_volumes_for_pooling(pools)
    except LimsError as e:
        sys.exit(e.message)



PLACEMENT_MAP_HEADER = """
    <head>
    <style>table, th, td {{border: 1px solid black; border-collapse: collapse;}}</style>
    <meta content="text/html; charset=UTF-8" http-equiv="Content-Type">
    <link href="../css/g/queue-print.css" rel="stylesheet" type="text/css" media="screen,print">
    <title>Pooling Map</title>
    </head>
    <body>
    <div id="header">
    <h1 class="title">{process_type}</h1>
    </div>
    <span>Created {date} </span>"""

PLATE_HEADER_SECTION = """
<tr>
<th class="group-header" colspan="10"><h2>Sample placement map: {container_name}</h2>
<table>
<tbody>
<tr>
<td class="group-field-label">{container_type}: </td>
<td class="group-field-value">{container_name}</td>
</tr>
<tr>
<td class="group-field-label">Container LIMS ID: </td>
<td class="group-field-value">{container_id}</td>
</tr>
</tbody>
</table>
<br>
</th>
</tr>"""

TABLE_HEADERS = """
<tr>
<th style="width: 7%;" class="">Project Name</th>
<th style="width: 7%;" class="">Sample/Pool Name</th>
<th style="width: 7%;" class="">Sample/Pool ID</th>
<th style="width: 7%;" class="">{container_source} Container</th>
<th style="width: 7%;" class="">{container_source} Well</th>
<th style="width: 7%;" class="">Dest. Well</th>
</tr>"""


TABLE_ROWS = """
<tr>
<td style="width: 7%;">{project_name}</td>
<td class="" style="width: 7%;">{sample_name}</td>
<td class="" style="width: 7%;">{sample_id}</td>
<td class="" style="width: 7%;">{container_name}</td>
<td class="" style="width: 7%;">{well}</td>
<td class="" style="width: 7%;">{dest_well}</td>
</tr>"""


VISUAL_PLACEMENT_MAP_HEADER = """
<table class="print-container-view">
<thead>
<tr>
<th>&nbsp;</th>
<th>1</th>
<th>2</th>
<th>3</th>
<th>4</th>
<th>5</th>
<th>6</th>
<th>7</th>
<th>8</th>
<th>9</th>
<th>10</th>
<th>11</th>
<th>12</th>
<th>13</th>
</tr>
</thead>"""

VISUAL_PLACEMENT_MAP_WELL = """
<td class="well" style="background-color: #CCC;">
Project : {project_name}<br>
{sample_type} Name : {sample_name}<br>
{sample_type} ID : {sample_id}<br>
{container_source} : {container_name}<br>
{well_source} : {well}<br>
{exta_udf_info}
</td>"""



#!/usr/bin/env python
import logging
import sys
from pathlib import Path
from typing import List

import click
from genologics.lims import Artifact

from cg_lims import options
from cg_lims.EPPs.files.hamilton.models import CovidPrepFileRow
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.files.manage_csv_files import build_csv, sort_csv
from cg_lims.get.artifacts import get_artifacts
from cg_lims.get.fields import get_index_well, get_artifact_well

LOG = logging.getLogger(__name__)

HEADERS = ["LIMS ID", "Sample Well", "Destination Well", "Index Well"]


def get_file_data_and_write(destination_artifacts: List[Artifact], file: str, pool: bool):
    """Making a sars cov 2 prep file for hamilton."""

    failed_samples = []
    file_rows = []
    for destination_artifact in destination_artifacts:
        source_artifacts: List[Artifact] = destination_artifact.input_artifact_list()
        for source_artifact in source_artifacts:
            try:
                row_data = CovidPrepFileRow(
                    lims_id=source_artifact.samples[0].id,
                    sample_well=get_artifact_well(source_artifact),
                    destination_well=get_artifact_well(destination_artifact),
                    index_well="-" if pool else get_index_well(artifact=destination_artifact),
                )
            except:
                failed_samples.append(source_artifact.id)
                continue

            row_data_dict = row_data.dict(by_alias=True)
            file_rows.append([row_data_dict[header] for header in HEADERS])

    build_csv(file=Path(file), rows=file_rows, headers=HEADERS)
    sort_csv(
        file=Path(file),
        columns=["Sample Well"],
        well_columns=["Sample Well"],
    )

    if failed_samples:
        raise MissingUDFsError(
            f"All samples were not added to the file. Udfs missing for samples: {', '.join(failed_samples)}"
        )


@click.command()
@options.file_placeholder(help="Hamilton Noramlization File")
@options.pooling_step()
@click.pass_context
def sars_cov2_prep_file(ctx: click.Context, file: str, pooling_step: bool):
    """Script to make a hamilton normalization file"""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    artifacts = get_artifacts(process=process, input=False)
    try:
        get_file_data_and_write(
            pool=pooling_step,
            destination_artifacts=artifacts,
            file=f"{file}-sars-cov2-hamilton-prep.txt",
        )
        click.echo("The file was successfully generated.")
    except LimsError as e:
        sys.exit(e.message)



PLACEMENT_MAP_HEADER = """
    <html>
    <head>
    <style>table, th, td {{border: 1px solid black; border-collapse: collapse;}}</style>
    <meta content="text/html; charset=UTF-8" http-equiv="Content-Type">
    <link href="../css/g/queue-print.css" rel="stylesheet" type="text/css" media="screen,print">
    <title>Pooling Map</title>
    </head>
    <body>
    <div id="header">
    <h1 class="title">{process_type}</h1>
    </div>
    Created: {date}
    """

POOL_HEADER = """
    <table class="group-contents">
    <thead>
    <tr><th class="group-header" colspan="10"><h2> {pool_name}</h2>
    <table>
    <tbody>
    <tr>
    <td class="group-field-label">Pool LIMS ID: </td>
    <td class="group-field-value">{pool_id}</td>
    </tr>
    <tr>
    <td class="group-field-label">Nr samples in pool: </td>
    <td class="group-field-value">{nr_samples}</td>
    </tr>
    {pools_information}
    </tbody>
    </table>
    <br>
    """

SAMPLE_COLUMN_HEADERS = """
    <tr>
    <th style="width: 7%;" class="">Sample Lims ID</th>
    <th style="width: 7%;" class="">Source Well</th>
    <th style="width: 7%;" class="">Source Container</th>
    {extra_sample_columns}
    </tr>
    """

SAMPLE_COLUMN_VALUES = """
    <tr>
    <td style="width: 7%;">{sample_id}</td>
    <td class="" style="width: 7%;">{source_well}</td>
    <td class="" style="width: 7%;">{source_container}</td>
    {extra_sample_values}
    </tr>
    """



#!/usr/bin/env python

import logging
import sys

import click
from genologics.lims import Lims

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.files.manage_csv_files import make_plate_file
from cg_lims.get.artifacts import get_artifacts, get_latest_artifact
from cg_lims.get.fields import get_index_well

LOG = logging.getLogger(__name__)


MINIMUM_AMOUNT = 10


def get_ligation_master_mix(amount):
    """Two different master mixes are used, A and B.
    Master mix B has less volume of "xGen CS Adapter-Tech Access"
    and is used for samples with DNA amount <25 ng.
    """

    if amount < 25:
        return "B"
    else:
        return "A"


def get_pcr_plate(amount):
    """Depending on amount of DNA the samples are run with different number
    of PCR-cycles. These are run in three different PCR-plates.
    The script sets the PCR-plate.
    """

    if amount < 50:
        return "Plate3"
    elif amount < 200:
        return "Plate2"
    else:
        return "Plate1"


def get_file_data_and_write(lims: Lims, amount_step: str, artifacts: list, file: str) -> dict:
    """Getting row data for hamilton file based on amount and reagent label"""

    failed_samples = []
    file_rows = {}
    for art in artifacts:
        sample_id = art.samples[0].id
        index_well = get_index_well(art)
        well = art.location[1].replace(":", "")
        amount_artifact = get_latest_artifact(lims, sample_id, amount_step)
        amount = amount_artifact.udf.get("Amount needed (ng)")

        if not amount:
            failed_samples.append(sample_id)
            continue

        amount = max(amount, MINIMUM_AMOUNT)

        ligation_master_mix = get_ligation_master_mix(amount)
        pcr_plate = get_pcr_plate(amount)

        file_rows[well] = [
            sample_id,
            well,
            ligation_master_mix,
            index_well,
            pcr_plate,
        ]
        art.udf["Ligation Master Mix"] = ligation_master_mix
        art.udf["PCR Plate"] = pcr_plate
        art.put()

    headers = [
        "LIMS ID",
        "Sample Well",
        "Ligation Master Mix",
        "Index Well",
        "PCR Plate",
    ]

    make_plate_file(file, file_rows, headers)

    if failed_samples:
        raise MissingUDFsError(
            f"Could not find udf: Amount needed (ng) for samples: {', '.join(failed_samples)}, from step {amount_step}."
        )


def resolve_file_extension(extension: str) -> str:
    """Appends a period symbol to extension string if none is present. Otherwise, returns given extension string"""
    if extension:
        if extension.startswith("."):
            return extension
        return f".{extension}"
    return ""


@click.command()
@options.process_type(help="Amount step name.")
@options.file_placeholder(help="Hamilton file.")
@options.file_extension()
@click.pass_context
def make_kapa_csv(ctx: click.Context, file: str, extension: str, process_type: str):
    """Script to make a csv file for hamilton. See AM doc #2125"""
    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    lims = ctx.obj["lims"]
    artifacts = get_artifacts(process=process, input=False)

    file_name = (
        f"{file}_{artifacts[0].container.name}_{process.type.name.replace(' ', '_')}"
        f"{resolve_file_extension(extension)}"
    )

    try:
        get_file_data_and_write(lims, process_type, artifacts, file_name)
        click.echo("The file was successfully generated.")
    except LimsError as e:
        sys.exit(e.message)
